/**
 * Insert text in colums spanning multiple lines
 *
 * Has special markers for inserting sequential list of numers
 *
 * @category  WeBuilder Plugin
 * @package   Column Insert
 * @author    Peter Klein <pmk@io.dk>
 * @copyright 2016 Peter Klein
 * @license   http://www.freebsd.org/copyright/license.html  BSD License
 * @version   1.01
 */

/**
 * [CLASS/FUNCTION INDEX of SCRIPT]
 *
 *     46   function ShowModal()
 *    213   function ShowHelpForm()
 *    256   function ColumnInsert()
 *    381   function LoadHistory(Sender)
 *    408   function SaveHistory(Sender)
 *    455   function GetSeqNumber(num, maxWidth, numberFormat)
 *    473   function IntToBin(num)
 *    496   function IntToRoman(num)
 *    519   function OnInstalled()
 *
 * TOTAL FUNCTIONS: 9
 * (This index is automatically created/updated by the WeBuilder plugin "DocBlock Comments")
 *
 */

/**
 * Location of history ini file
 *
 * @var historyFile
 */
var historyFile = Script.GetPath + "history.ini";

/**
 *
 * Create the Modal GUI Form
 *
 * @return      void
 *
 */
function ShowModal() {
    var mleft = 14,
        mtop = 14,
        vspace = 32,
        hdiv = 10,
        lspace = 4,
        bspace = 16,
        topPos = 0,
        leftPos = 0,
        gTopPos = 0,
        gLeftPos = 0;
        margin = 6;

    var SL = new TStringList;
    if (Editor.SelText != "") {
        SL.Text = Editor.SelText;
    }
    else {
        SL.Text = Editor.Text;
    }
    var lineCount = SL.Count;
    delete SL;

    var modalForm = new TForm(WeBuilder);
    modalForm.ClientWidth = 410;
    modalForm.ClientHeight = 155;
    modalForm.Position = poScreenCenter;
    modalForm.BorderStyle = bsSingle; //disable dialog resizing
    modalForm.BorderIcons = biSystemMenu; //remove maximize & minimize buttons
    modalForm.Caption = "Column Insert";
    modalForm.Color = clBtnFace;
    modalForm.Font.Color = clWindowText;
    modalForm.Font.Height = -13;
    modalForm.Font.Name = "Tahoma";

    var groupWidth = modalForm.ClientWidth - margin - margin - 8;
    gTopPos = margin;

    groupBox1 = new TGroupBox(WeBuilder);
    groupBox1.Parent = modalForm;
    groupBox1.Caption = "Column Insert";
    groupBox1.Anchors = akLeft + akTop;
    groupBox1.Width = groupWidth;
    groupBox1.Left = margin;
    groupBox1.Top = gTopPos;

    topPos = mtop + mtop;
    leftPos = mleft;

    label1 = new TLabel(WeBuilder);
    label1.Parent = groupBox1;
    label1.Anchors = akLeft + akTop;
    label1.Alignment = taRightJustify;
    label1.ShowHint = true;
    label1.Hint = "Text to Insert";
    label1.Caption = "Text to Insert:";
    label1.SetBounds(leftPos, topPos + 3, 82, 17);

    leftPos += label1.Width + lspace;

    editBox1 = new TComboBox(WeBuilder);
    editBox1.Parent = groupBox1;
    editBox1.Anchors = akLeft  + akTop;
    editBox1.TabOrder = 0;
    editBox1.Text = "";
    editBox1.ShowHint = true;
    editBox1.Hint = "Text to Insert";
    editBox1.SetBounds( leftPos, topPos, 265, 25);
    // Load editBox1 history
    LoadHistory(editBox1);

    leftPos += editBox1.Width + lspace;

    helpButton = new TBitBtn(WeBuilder);
    helpButton.Parent = groupBox1;
    helpButton.Glyph.LoadFromFile(Script.Path + "/help.bmp");
    helpButton.OnClick = "ShowHelpForm";
    helpButton.ShowHint = TRUE;
    helpButton.Hint = "Show Help";
    helpButton.SetBounds( leftPos, topPos, 25, 25);

    topPos += vspace;
    leftPos = mleft;

    label2 = new TLabel(WeBuilder);
    label2.Parent = groupBox1;
    label2.Alignment = taRightJustify;
    label2.Anchors = akLeft + akTop;
    label2.ShowHint = true;
    label2.Hint = "At Position";
    label2.Caption = "At Position:";
    label2.SetBounds(leftPos, topPos + 3, 82, 17);

    leftPos += label2.Width + lspace;

    editBox2 = new TEdit(WeBuilder);
    editBox2.Parent = groupBox1;
    editBox2.Anchors = akLeft  + akTop;
    editBox2.Alignment = taRightJustify;
    editBox2.TabOrder = 1;
    editBox2.ShowHint = true;
    editBox2.Hint = "At Position";
    editBox2.NumbersOnly = true;
    editBox2.Text = _t(Editor.Selection.SelStartColReal);
    editBox2.SetBounds(leftPos, topPos, 50, 25);

    leftPos += editBox2.Width + hdiv + hdiv;

    // Dummy panel to display hint when Checkbox is disbaled
    pnlReplace = new TPanel(WeBuilder);
    pnlReplace.Parent = groupBox1;
    pnlReplace.ShowHint = true;
    pnlReplace.Hint = "Replace instead of insert. Only available if selection is made in Block Select Mode.";
    pnlReplace.SetBounds(leftPos, topPos, 72, 23);

    chkReplace = new TCheckBox(WeBuilder);
    chkReplace.Parent = groupBox1;
    chkReplace.Alignment = taLeftJustify;
    chkReplace.ShowHint = true;
    chkReplace.Hint = "Replace instead of insert.";
    chkReplace.Caption = "Replace:";
    chkReplace.Checked = Actions.IsChecked("ActBlockSelect");
    chkReplace.Enabled = Actions.IsChecked("ActBlockSelect");
    chkReplace.SetBounds(leftPos, topPos, 72, 23);

    groupBox1.Height = topPos + mtop + mtop + 13;
    gTopPos += groupBox1.Height + margin + 5;

    leftPos = mleft;

    label3 = new TLabel(WeBuilder);
    label3.Parent = modalForm;
    label3.Anchors = akLeft + akTop;
    label3.Font.Style = fsItalic;
    label3.Caption = _t(lineCount) + " lines will be modified";
    label3.SetBounds(leftPos, gTopPos + 8, 160, 17);

    // Cancel button object
    btnCancel = new TButton(WeBuilder);
    btnCancel.Parent = modalForm;
    btnCancel.Caption = "Cancel";
    btnCancel.Cancel = True;
    btnCancel.ModalResult = mrCancel;
    btnCancel.SetBounds(modalForm.Width - margin - 112 - 98 - bspace, gTopPos, 98, 33);

    // OK button object
    btnOk = new TButton(WeBuilder);
    btnOk.Parent = modalForm;
    btnOk.Caption = "OK";
    btnOk.Default = True;
    btnOk.ModalResult = mrOk;
    btnOk.SetBounds(modalForm.Width - margin - 112, gTopPos, 98, 33);

    var modalResult = modalForm.ShowModal;
    if (modalResult == mrOK) {
        ColumnInsert();
    }

    // Remove Modal object
    delete modalForm;
}

/**
 * Show Help text in seperate window
 *
 * @return void
 */
function ShowHelpForm() {
    var groupMargin = 12,lf = chr(13);

    modalFormHelp = new TForm(WeBuilder);
    modalFormHelp.ClientWidth = 476;
    modalFormHelp.ClientHeight = 332;
    modalFormHelp.Position = poScreenCenter;
    modalFormHelp.BorderStyle = bsSingle; //disable dialog resizing
    modalFormHelp.BorderIcons = biSystemMenu; //remove maximize & minimize buttons
    modalFormHelp.Caption = "Column Insert Help";
    modalFormHelp.Color = clBtnFace;
    modalFormHelp.Font.Color = clWindowText;
    modalFormHelp.Font.Height = -13;
    modalFormHelp.Font.Name = "Tahoma";

    groupBoxHelp = new TGroupBox(modalFormHelp);
    groupBoxHelp.Parent = modalFormHelp;
    groupBoxHelp.SetBounds(groupMargin, groupMargin, modalFormHelp.ClientWidth - groupMargin*2 , modalFormHelp.ClientHeight - groupMargin *2 - 60);
    groupBoxHelp.Caption = "Special markers to insert incremental list of numbers";

    labelHelp = new TLabel(groupBoxHelp);
    labelHelp.Parent = modalFormHelp;
    labelHelp.Caption = "%d - number inserted in decimal format" + lf + "%h - number inserted in hexadecimal format (lower case)" + lf + "%H - number inserted in hexadecimal format (upper case)" + lf + "%b - number inserted in binary format"  + lf + "%r - number inserted in roman numerial format" + lf + lf + "As default the number will start at 1, but you can change that by adding" + lf + "a (decimal) number before the marker letter. Like:" + lf + lf + "%6d - numbers inserted in decimal format, starting at 6." + lf + "%16h - numbers inserted in hexadecimal format, starting at 16 (hex 0f)." + lf + lf + "Note: Only the first marker found will be processed.";
    labelHelp.SetBounds(20, 42, 410, 212);

    // Close button object
    btnClose = new TButton(modalFormHelp);
    btnClose.Parent = modalFormHelp;
    btnClose.Caption = "Close";
    btnClose.Default = True;
    btnClose.ModalResult = mrCancel;
    btnClose.SetBounds(modalFormHelp.Width - 75 - 16, 292, 75, 25);

    var modalResult = modalFormHelp.ShowModal;
    // Remove Modal object
    delete modalFormHelp;
}

/**
 * Insert column at position
 *
 * @return void
 */
function ColumnInsert() {

	var insertText = editBox1.Text;

	// Nothing to insert, so we return
	if (insertText == "") return;

    // Test if in Block Select Mode
    var blockMode = Actions.IsChecked("ActBlockSelect");

    if (blockMode) {
        // Turn Block Select Mode off while doing the insert
        Actions.Execute("ActBlockSelect");
    }

	var insertPos =  StrToInt("0" + editBox2.Text),
        uSel = Editor.Selection,
        SL = new TStringList,
        startCol = 0,
        endCol = 0,
        replacelength = 0;

	if (Editor.SelText != "") {
  	    // we got a selection, so we do insert only on selected lines
        // Save start and end column of selection
        startCol = uSel.SelStartCol;
        endCol = uSel.SelEndCol;

        if (chkReplace.Checked) {
            // Replace Mode
            replaceLength = endCol - startCol;
        }

        // Move selection to start and end of line
        uSel.SelStartCol = 0;
        uSel.SelEndCol = Length(editor.Lines[uSel.SelEndLine]);
        Editor.Selection = uSel;

	    SL.Text = Editor.SelText;
	}
	else {
		// No selection, so we do insert on all lines
	    SL.Text = Editor.Text;
	}

	// Does the insert text contain any special markers?
	var replaceMarker = "", startCount, numberFormat, maxWidth;
	if (RegexMatchAll(insertText, "%(\\d*)([dhbr])", true, matches, positions) == true) {
		if (Length(matches)) {
			// We only want the first marker found
			startCount = _v(matches, [0, 1]);
			numberFormat = _v(matches, [0, 2]);
			replaceMarker = "%" + startCount + numberFormat;
			if (startCount == "") startCount = "1";
			startCount = StrToInt(startCount);
			// find width of largest number
            // Doesn't work on Roman numbers as they vary in size
			maxWidth = Length(GetSeqNumber(SL.Count + startCount, "", numberFormat));
		}
	}

	var i,line;
	for (i=0;i<SL.Count;i++) {
		line = SL[i];
		// Make sure line is long enough to insert text at the correct position
		if (Length(line) < insertPos) {
			line = Format("%-" + _t(insertPos) + "s",[line]);
		}

		tempInsertText = insertText;
		if (replaceMarker != "") {
			tempInsertText = Replace(insertText, replaceMarker, GetSeqNumber(startCount + i, maxWidth, numberFormat));
		}

		// replace line with new line
		SL[i] = Copy(line, 0, insertPos) + tempInsertText + Copy(line, insertPos + replaceLength + 1, Length(line));

	}

	Editor.BeginEditing;

    // Get rid of trailing linefeeds before applying insert
    var tmpText = RegexReplace(SL.Text, "(\\r|\\n|\\r\\n)$", "", false);

	if (Trim(Editor.SelText) != "") {
  	    // we got a selection, so we do insert only on selected lines
	    Editor.SelText = tmpText;
        // Restore selection.
        uSel.SelStartCol = startCol;
        if (chkReplace.Checked) {
            // Replace Mode
            uSel.SelEndCol = startCol + Length(tempInsertText);
        }
        else {
            uSel.SelEndCol = endCol + Length(tempInsertText);
        }

        Editor.Selection = uSel;
	}
	else {
		// No selection, so we do insert on all lines
	    Editor.Text = tmpText;
	}

    if (blockMode) {
        // Turn Block Select Mode back on if it was active before insert
        Actions.Execute("ActBlockSelect")
    }

	Editor.EndEditing;

    // Save editBox1 history
    SaveHistory(editBox1);

	// Object Cleanup
    delete SL;
}

/**
 * Load editbox history
 *
 * @param  TComboBox   Sender
 *
 * @return void
 */
function LoadHistory(Sender) {
    var val, i;
    var ini = new TIniFile(historyFile);
    var SL = new TStringList;

    if (ini.SectionExists("History")) {
        ini.ReadSectionValues("History", SL);
        for (i=0;i<SL.Count;i++) {
            val = RegexReplace(SL[i],"^\\d*=", "", true);
            if (val!="") {
                Sender.Items.Add(val);
            }
        }
    }

	// Object Cleanup
    delete SL;
    delete ini;
}

/**
 * Save editbox history
 *
 * @param  TComboBox   Sender
 *
 * @return void
 */
function SaveHistory(Sender) {
    var ini = new TIniFile(historyFile);
    var SL = new TStringList;
    var i, j, dupe;

    // Add the current insert text as 1st item
    SL.Add(Trim(Sender.Text));

    // Add existing history items
    for (i = 0; i < Length(Sender.Items); i++) {
        // Skip duplicates. (TStringList dupe functionality only works when list is sorted)
        dupe = false;
        for (j = 0; j < SL.Count; j++) {
            if (Sender.Items[i] == SL[j]) {
                dupe = true;
            }
        }
        if (!dupe) {
            SL.Add(Sender.Items[i]);
        }
    }

    if (ini.SectionExists("History")) {
        // Clear section
        ini.EraseSection("History");

        // Save updated history
        for (i = 0; (i < SL.Count) && (i < 10); i++) {
            if (SL[i] != "") {
                ini.WriteString("History", i, SL[i]);
            }
        }
    }

	// Object Cleanup
    delete SL;
    delete ini;
}

/**
 * Format sequential numbering
 *
 * @param  int   num
 * @param  int/string   maxWidth number or empty string
 *
 * @return string
 */
function GetSeqNumber(num, maxWidth, numberFormat) {
	switch (numberFormat) {
		case "b": return Format("%" + _t(maxWidth) + "." + _t(maxWidth) + "d",[IntToBin(num)]);
		case "h": return Lowercase(Format("%" + _t(maxWidth) + "." + _t(maxWidth) + "x",[num]));
		case "H": return Uppercase(Format("%" + _t(maxWidth) + "." + _t(maxWidth) + "x",[num]));
		case "r": return Format("%s",[IntToRoman(num)]);
		default: return Format("%" + _t(maxWidth) + "d",[num]);
	}
}


/**
 * Convert integer decimal number to binary number format
 *
 * @param  integer   num - Then number to convert
 *
 * @return integer
 */
function IntToBin(num) {
    var digits = 8;
    var res = "";
    while (num > 0) {
        if (num && 1) {
            res = "1" + res;
        }
        else {
            res = "0" + res;
        }
        digits--;
        num = num >> 1;
    }
    return StrToInt(res);
}

/**
 * Convert integer arabic number to roman number format
 *
 * @param  integer   num - Then number to convert
 *
 * @return string
 */
function IntToRoman(num) {
    var i, val = num, res = "";
    var arabics = [0,1,4,5,9,10,40,50,90,100,400,500,900,1000];
    var romans = [" ","I","IV","V","IX","X","XL","L","XC","C","CD","D","CM","M"];
    if (num == 0) {
        res = "nulla";
    }
    else {
        for (i=13;i;i--) {
            while (val >= arabics[i]) {
                val = val - arabics[i];
                res = res + romans[i];
            }
        }
    }
    return res;
}

/**
 * Show info when plugin is installed
 *
 * @return void
 */
function OnInstalled() {
  alert("Column Insert 1.01 by Peter Klein installed sucessfully!");
}

Script.ConnectSignal("installed", "OnInstalled");
var bmp = new TBitmap;
var icon = LoadFileToBitmap(Script.Path + "columninsert.png", bmp);
var act = Script.RegisterDocumentAction("", "Column Insert", "", "ShowModal");
Actions.SetIcon(act, bmp);
delete bmp;